/*
 * (C) Copyright 2008-2009
 * Max Tretene, ACube Systems Srl. mtretene@acube-systems.com.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <ppc4xx.h>
#include <asm/processor.h>
#include <asm/mmu.h>
#include <i2c.h>
#include <part.h>
#include <command.h>
#include <usb.h>
#include <video_fb.h>
#include <sm501.h>
#include "spd_sdram.h"
#include "Sam440ep.h"
#include "../common/vesa.h"

#define SAMLOGO
#ifdef SAMLOGO
#include "../common/logo_acube.h"
#else
#define LOGO_WIDTH 176
#define LOGO_HEIGHT 48
unsigned char logo_acube[LOGO_WIDTH * LOGO_HEIGHT] = { 0 };
#endif

DECLARE_GLOBAL_DATA_PTR;

#define SAM_DEBUG

#ifdef	SAM_DEBUG
#define PRINTF(fmt,args...)		printf (fmt ,##args)
#else
#define PRINTF(fmt,args...)
#endif

#ifndef CFG_NO_FLASH
extern flash_info_t flash_info[CFG_MAX_FLASH_BANKS]; /* info for FLASH chips	*/
#endif

extern int console_col; /* cursor col */
extern int console_row; /* cursor row */

struct FrameBufferInfo *fbi = NULL;

extern int pci_skip_dev(struct pci_controller *hose, pci_dev_t dev);

int board_early_init_f(void)
{
	register uint reg;

	/*--------------------------------------------------------------------
	 * Setup the external bus controller/chip selects
	 *-------------------------------------------------------------------*/

	mtdcr(ebccfga, xbcfg);
	reg = mfdcr(ebccfgd);
	mtdcr(ebccfgd, reg | 0x04000000);	// Set ATC

	mtebc(pb0ap, 0x03017240);	// CS0 - EEProm
	mtebc(pb0cr, 0xfff18000);	// BAS=0xfff 1MB ro 8-bit

	mtebc(pb1ap, 0x03017300);	// CS1 - Disk On Chip
	mtebc(pb1cr, 0xf0088000);	// BAS=0xf00 8MB r/w 8-bit

	mtebc(pb2ap, 0x00000000);	// CS2 - SmartMedia
	mtebc(pb2cr, 0x00000000);	//

	mtebc(pb3ap, 0x03017240);	// CS3 - FPGA
	mtebc(pb3cr, 0xfe05a000);	// BAS=0xfe000000 4MB r/w 16-bit

	mtebc(pb4ap, 0x00000000);
	mtebc(pb4cr, 0x00000000);

	mtebc(pb5ap, 0x00000000);
	mtebc(pb5cr, 0x00000000);

	/*--------------------------------------------------------------------
	 * Setup the GPIO pins
	 *-------------------------------------------------------------------*/

	/*CPLD cs */
	/*setup Address lines for flash size 64Meg. */
	out32(GPIO0_OSRL, in32(GPIO0_OSRL) | 0x50010000);
	out32(GPIO0_TSRL, in32(GPIO0_TSRL) | 0x50010000);
	out32(GPIO0_ISR1L, in32(GPIO0_ISR1L) | 0x50000000);

	/*setup emac */
	out32(GPIO0_TCR, in32(GPIO0_TCR) | 0xC080);
	out32(GPIO0_TSRL, in32(GPIO0_TSRL) | 0x40);
	out32(GPIO0_ISR1L, in32(GPIO0_ISR1L) | 0x55);
	out32(GPIO0_OSRH, in32(GPIO0_OSRH) | 0x50004000);
	out32(GPIO0_ISR1H, in32(GPIO0_ISR1H) | 0x00440000);

	/*UART1 */
	out32(GPIO1_TCR, in32(GPIO1_TCR) | 0x02000000);
	out32(GPIO1_OSRL, in32(GPIO1_OSRL) | 0x00080000);
	out32(GPIO1_ISR2L, in32(GPIO1_ISR2L) | 0x00010000);

	/* external interrupts IRQ0...4 and IRQ6...7 */
	out32(GPIO1_TCR, in32(GPIO1_TCR) & ~0x00fe0000);
	out32(GPIO1_TSRL, in32(GPIO1_TSRL) & ~0x0000fffc);
	out32(GPIO1_ISR1L, in32(GPIO1_ISR1L) | 0x00005554);

	/*--------------------------------------------------------------------
	 * Setup the interrupt controller polarities, triggers, etc.
	 *-------------------------------------------------------------------*/
	mtdcr(uic0sr, 0xffffffff);	/* clear all */
	mtdcr(uic0er, 0x00000000);	/* disable all */
	mtdcr(uic0cr, 0x00000009);	/* ATI & UIC1 crit are critical */
	mtdcr(uic0pr, 0xffbffe03);
	mtdcr(uic0tr, 0x00000008);	/* per ref-board manual */
	mtdcr(uic0vr, 0x00000001);	/* int31 highest, base=0x000 */
	mtdcr(uic0sr, 0xffffffff);	/* clear all */

	mtdcr(uic1sr, 0xffffffff);	/* clear all */
	mtdcr(uic1er, 0x00000000);	/* disable all */
	mtdcr(uic1cr, 0x00000000);	/* all non-critical */
	mtdcr(uic1pr, 0xffffe6af);
	mtdcr(uic1tr, 0x06000140);	/* per ref-board manual */
	mtdcr(uic1vr, 0x00000001);	/* int31 highest, base=0x000 */
	mtdcr(uic1sr, 0xffffffff);	/* clear all */

	/*--------------------------------------------------------------------
	 * Setup other serial configuration
	 *-------------------------------------------------------------------*/
	mfsdr(sdr_pci0, reg);
	mtsdr(sdr_pci0, 0x80000000 | reg);	/* PCI arbiter enabled */
	mtsdr(sdr_pfc0, 0x00003e00);	/* Pin function */
	mtsdr(sdr_pfc1, 0x00088000);	/* Pin function: UART0 has 8 pins */

	return 0;
}

/*
 * Disk On Chip (DOC) 2000 initialization.
 * The DOC lives in the CS2* space
 */
#if (CONFIG_COMMANDS & CFG_CMD_DOC)
#define CFG_DOC_BASE 0xF0000000

extern void doc_probe (ulong physadr);

void doc_init (void)
{
	// not mplemented

	return;
}
#endif

int misc_init_r (void)
{
	return 0;
}

int checkboard(void)
{
	sys_info_t sysinfo;

	get_sys_info(&sysinfo);
	puts ("Board: Sam440ep\n");

	return (0);
}

// 512 Mb DDR400 - DIMM 100 Module
/*
static void init_spd_array_512(unsigned char *cfg_simulate_spd_eeprom)
{
	cfg_simulate_spd_eeprom[0]     = 0x80;    // number of SPD bytes used: 128
	cfg_simulate_spd_eeprom[1]     = 0x08;    // total number bytes in SPD device = 256
	cfg_simulate_spd_eeprom[2]     = 0x07;    // DDR ram
	cfg_simulate_spd_eeprom[3]     = 0x0D;    // num Row Addr: 13

	cfg_simulate_spd_eeprom[4]     = 0x0B;    // numColAddr: 11
	cfg_simulate_spd_eeprom[5]     = 0x02;    // numBanks: 1
	cfg_simulate_spd_eeprom[6]     = 0x20;    // Module data width: 32 bits
	cfg_simulate_spd_eeprom[7]     = 0x00;    // Module data width continued: +0

	cfg_simulate_spd_eeprom[8]     = 0x04;    // 2.5 Volt
	cfg_simulate_spd_eeprom[9]     = 0x60;    // SDRAM Cycle Time (cas latency 2.5) = 7.5 ns
	cfg_simulate_spd_eeprom[10]    = 0x70;
	cfg_simulate_spd_eeprom[11]    = 0x00;    // ECC ON : 02 OFF : 00

	cfg_simulate_spd_eeprom[12]    = 0x82;    // refresh Rate Type: Normal (15.625us) + Self refresh
	cfg_simulate_spd_eeprom[13]    = 0x08;
	cfg_simulate_spd_eeprom[14]    = 0x00;
	cfg_simulate_spd_eeprom[15]    = 0x01;    // wcsbc = 1

	cfg_simulate_spd_eeprom[16]    = 0x0E;
	cfg_simulate_spd_eeprom[17]    = 0x04;
	cfg_simulate_spd_eeprom[18]    = 0x0C;    // casBit (2,2.5)
	cfg_simulate_spd_eeprom[19]    = 0x01;

	cfg_simulate_spd_eeprom[20]    = 0x02;
	cfg_simulate_spd_eeprom[21]    = 0x20;    // not registered: 0  registered : 0x02
	cfg_simulate_spd_eeprom[22]    = 0xC0;
	cfg_simulate_spd_eeprom[23]    = 0x75;    // SDRAM Cycle Time (cas latency 2) = 10 ns

	cfg_simulate_spd_eeprom[24]    = 0x70;
	cfg_simulate_spd_eeprom[25]    = 0x00;    // SDRAM Cycle Time (cas latency 1.5) = N.A
	cfg_simulate_spd_eeprom[26]    = 0x00;
	cfg_simulate_spd_eeprom[27]    = 0x48;    // tRpNs = 20 ns

	cfg_simulate_spd_eeprom[28]    = 0x30;
	cfg_simulate_spd_eeprom[29]    = 0x48;    // tRcdNs = 20 ns
	cfg_simulate_spd_eeprom[30]    = 0x2a;    // tRasNs
	cfg_simulate_spd_eeprom[31]    = 0x40;    // bankSizeID: 0x20 = 128MB 0x40 = 256MB
}
*/
/*************************************************************************
 *  long int initdram
 *
 ************************************************************************/
long int initdram(int board)
{
	long dram_size = 0;

	unsigned char cfg_simulate_spd_eeprom[32] = {
		0x80,0x08,0x07,0x0D,
		0x0B,0x01,0x20,0x00,
		0x04,0x60,0x70,0x00,
		0x82,0x08,0x00,0x01,
		0x0E,0x04,0x0c,0x01,
		0x02,0x20,0xc0,0x75,
		0x70,0x00,0x00,0x48,
		0x30,0x48,0x2a,0x40 };

	dram_size = spd_sdram(cfg_simulate_spd_eeprom);

	return dram_size;
}

#if defined(CFG_DRAM_TEST)
int testdram(void)
{
	unsigned long *mem = (unsigned long *)0;
	const unsigned long kend = (1024 / sizeof(unsigned long));
	unsigned long k, n;
	unsigned long value;
	int error = 0;

	char *p = getenv("testram");
	if (p != NULL) {
		mtmsr(0);

		for (k = 0; k < gd->ram_size/1024;
		     ++k, mem += (1024 / sizeof(unsigned long))) {
			if ((k & 1023) == 0) {
				printf("Testing: %3d MB\r", k / 1024);
			}

			for (n = 0; n < kend; ++n) {
				value = (n << 24 | n << 16 | n << 8 | n);
				mem[n] = value;
			}

			for (n = 0; n < kend; ++n) {
				value = (n << 24 | n << 16 | n << 8 | n);
				if (mem[n] != value) {
					printf("SDRAM test fails at %08x, wrote %08x read %08x\n",
					       (uint) & mem[n],value,mem[n]);
					error = 1;
				}
			}
		}

		if (error == 0) puts ("SDRAM test OK\n");
		return error;
	}
	else return 0;
}
#endif

/*************************************************************************
 *  pci_pre_init
 *
 *  This routine is called just prior to registering the hose and gives
 *  the board the opportunity to check things. Returning a value of zero
 *  indicates that things are bad & PCI initialization should be aborted.
 *
 *	Different boards may wish to customize the pci controller structure
 *	(add regions, override default access routines, etc) or perform
 *	certain pre-initialization actions.
 *
 ************************************************************************/
#if defined(CONFIG_PCI) && defined(CFG_PCI_PRE_INIT)
int pci_pre_init(struct pci_controller *hose)
{
	unsigned long addr;

	/*-------------------------------------------------------------------------+
	  | Set priority for all PLB3 devices to 0.
	  | Set PLB3 arbiter to fair mode.
	  +-------------------------------------------------------------------------*/
	mfsdr(sdr_amp1, addr);
	mtsdr(sdr_amp1, (addr & 0x000000FF) | 0x0000FF00);
	addr = mfdcr(plb3_acr);
	mtdcr(plb3_acr, addr | 0x88000000); // high load enable - was 0x80000000

	/*-------------------------------------------------------------------------+
	  | Set priority for all PLB4 devices to 0.
	  +-------------------------------------------------------------------------*/
	mfsdr(sdr_amp0, addr);
	mtsdr(sdr_amp0, (addr & 0x000000FF) | 0x0000FF00);
	addr = mfdcr(plb4_acr) | 0xa0000000;	/* Was 0x8---- */
	mtdcr(plb4_acr, addr);

	/*-------------------------------------------------------------------------+
	  | Set Nebula PLB4 arbiter to fair mode.
	  +-------------------------------------------------------------------------*/
	/* Segment0 */
	addr = (mfdcr(plb0_acr) & ~plb0_acr_ppm_mask) | plb0_acr_ppm_fair;
	addr = (addr & ~plb0_acr_hbu_mask) | plb0_acr_hbu_enabled;
	addr = (addr & ~plb0_acr_rdp_mask) | plb0_acr_rdp_4deep;
	addr = (addr & ~plb0_acr_wrp_mask) | plb0_acr_wrp_2deep;
	mtdcr(plb0_acr, addr);

	/* Segment1 */
	addr = (mfdcr(plb1_acr) & ~plb1_acr_ppm_mask) | plb1_acr_ppm_fair;
	addr = (addr & ~plb1_acr_hbu_mask) | plb1_acr_hbu_enabled;
	addr = (addr & ~plb1_acr_rdp_mask) | plb1_acr_rdp_4deep;
	addr = (addr & ~plb1_acr_wrp_mask) | plb1_acr_wrp_2deep;
	mtdcr(plb1_acr, addr);

	return 1;
}
#endif				/* defined(CONFIG_PCI) && defined(CFG_PCI_PRE_INIT) */

/*************************************************************************
 *  pci_target_init
 *
 *	The bootstrap configuration provides default settings for the pci
 *	inbound map (PIM). But the bootstrap config choices are limited and
 *	may not be sufficient for a given board.
 *
 ************************************************************************/
#if defined(CONFIG_PCI) && defined(CFG_PCI_TARGET_INIT)
void pci_target_init(struct pci_controller *hose)
{
	/*--------------------------------------------------------------------------+
	 * Set up Direct MMIO registers
	 *--------------------------------------------------------------------------*/
	/*--------------------------------------------------------------------------+
	  | PowerPC440 EP PCI Master configuration.
	  | Map one 1Gig range of PLB/processor addresses to PCI memory space.
	  |   PLB address 0xA0000000-0xDFFFFFFF ==> PCI address 0xA0000000-0xDFFFFFFF
	  |   Use byte reversed out routines to handle endianess.
	  | Make this region non-prefetchable.
	  +--------------------------------------------------------------------------*/
	out32r(PCIX0_PMM0MA,    0);                /* PMM0 Mask/Attribute - disabled b4 setting */
	out32r(PCIX0_PMM0LA,    CFG_PCI_MEMBASE);  /* PMM0 Local Address */
	out32r(PCIX0_PMM0PCILA, CFG_PCI_MEMBASE);  /* PMM0 PCI Low Address */
	out32r(PCIX0_PMM0PCIHA, 0x00000000);       /* PMM0 PCI High Address */
	out32r(PCIX0_PMM0MA,    0xE0000001);       /* 512M + No prefetching, and enable region */

	out32r(PCIX0_PMM1MA,    0);                /* PMM0 Mask/Attribute - disabled b4 setting */
	out32r(PCIX0_PMM1LA,    CFG_PCI_MEMBASE2); /* PMM0 Local Address */
	out32r(PCIX0_PMM1PCILA, CFG_PCI_MEMBASE2); /* PMM0 PCI Low Address */
	out32r(PCIX0_PMM1PCIHA, 0x00000000);       /* PMM0 PCI High Address */
	out32r(PCIX0_PMM1MA,    0xE0000001);       /* 512M + No prefetching, and enable region */

	out32r(PCIX0_PMM2MA,    0);                /* PMM2 Mask/Attribute - disabled */
	out32r(PCIX0_PMM2LA,    0);                /* PMM2 Local Address */
	out32r(PCIX0_PMM2PCILA, 0);                /* PMM2 PCI Low Address */
	out32r(PCIX0_PMM2PCIHA, 0);                /* PMM2 PCI High Address */

	out32r(PCIX0_PTM1MS, 0x00000001);          /* Memory Size/Attribute */
	out32r(PCIX0_PTM1LA, 0);                   /* Local Addr. Reg */
	out32r(PCIX0_PTM2MS, 0);                   /* Memory Size/Attribute */
	out32r(PCIX0_PTM2LA, 0);                   /* Local Addr. Reg */

	/*--------------------------------------------------------------------------+
	 * Set up Configuration registers
	 *--------------------------------------------------------------------------*/

	/* Program the board's subsystem id/vendor id */
	pci_write_config_word(0, PCI_SUBSYSTEM_VENDOR_ID,
			      CFG_PCI_SUBSYS_VENDORID);
	pci_write_config_word(0, PCI_SUBSYSTEM_ID, CFG_PCI_SUBSYS_ID);

	/* Configure command register as bus master */
	pci_write_config_word(0, PCI_COMMAND, PCI_COMMAND_MASTER);

	/* 240nS PCI clock */
	pci_write_config_byte(0, PCI_LATENCY_TIMER, 0x0);

	/* No error reporting */
	pci_write_config_word(0, PCI_ERREN, 0);

	pci_write_config_dword(0, PCI_BRDGOPT2, 0x00000101);

}
#endif				/* defined(CONFIG_PCI) && defined(CFG_PCI_TARGET_INIT) */

/*************************************************************************
 *  pci_master_init
 *
 ************************************************************************/
#if defined(CONFIG_PCI) && defined(CFG_PCI_MASTER_INIT)
void pci_master_init(struct pci_controller *hose)
{
	unsigned short temp_short;

	/*--------------------------------------------------------------------------+
	  | Write the PowerPC440 EP PCI Configuration regs.
	  |   Enable PowerPC440 EP to be a master on the PCI bus (PMM).
	  |   Enable PowerPC440 EP to act as a PCI memory target (PTM).
	  +--------------------------------------------------------------------------*/
	pci_read_config_word(0, PCI_COMMAND, &temp_short);
	pci_write_config_word(0, PCI_COMMAND,
			      temp_short | PCI_COMMAND_MASTER |
			      PCI_COMMAND_MEMORY);
}
#endif				/* defined(CONFIG_PCI) && defined(CFG_PCI_MASTER_INIT) */

/*************************************************************************
 *  is_pci_host
 *
 *	This routine is called to determine if a pci scan should be
 *	performed. With various hardware environments (especially cPCI and
 *	PPMC) it's insufficient to depend on the state of the arbiter enable
 *	bit in the strap register, or generic host/adapter assumptions.
 *
 *	Rather than hard-code a bad assumption in the general 440 code, the
 *	440 pci code requires the board to decide at runtime.
 *
 *	Return 0 for adapter mode, non-zero for host (monarch) mode.
 *
 *
 ************************************************************************/
#if defined(CONFIG_PCI)
int is_pci_host(struct pci_controller *hose)
{
	/* Sam440ep is always configured as host. */
	return (1);
}
#endif				/* defined(CONFIG_PCI) */

/*************************************************************************
 *  hw_watchdog_reset
 *
 *	This routine is called to reset (keep alive) the watchdog timer
 *
 ************************************************************************/
#if defined(CONFIG_HW_WATCHDOG)
void hw_watchdog_reset(void)
{

}
#endif

void pciauto_setup_device_mem(struct pci_controller *hose,
			  pci_dev_t dev, int bars_num,
			  struct pci_region *mem,
			  struct pci_region *prefetch,
			  struct pci_region *io,
			  u32 bar_size_lower,
			  u32 bar_size_upper)
{
	unsigned int bar_response, bar_back;
	u32 bar_value;
	u32 bar_size;
	struct pci_region *bar_res;
	int bar, bar_nr = 0;
	int found_mem64 = 0;

	for (bar = PCI_BASE_ADDRESS_0; bar < PCI_BASE_ADDRESS_0 + (bars_num*4); bar += 4) {
		/* Tickle the BAR and get the response */
		pci_hose_read_config_dword(hose, dev, bar, &bar_back);
		pci_hose_write_config_dword(hose, dev, bar, 0xffffffff);
		pci_hose_read_config_dword(hose, dev, bar, &bar_response);
		pci_hose_write_config_dword(hose, dev, bar, bar_back);

		/* If BAR is not implemented go to the next BAR */
		if (!bar_response)
			continue;

		found_mem64 = 0;

		/* Check the BAR type and set our address mask */
		if ( ! (bar_response & PCI_BASE_ADDRESS_SPACE)) {
			if ( (bar_response & PCI_BASE_ADDRESS_MEM_TYPE_MASK) ==
				 PCI_BASE_ADDRESS_MEM_TYPE_64) {
				u32 bar_response_upper;
				u64 bar64;
				pci_hose_write_config_dword(hose, dev, bar+4, 0xffffffff);
				pci_hose_read_config_dword(hose, dev, bar+4, &bar_response_upper);

				bar64 = ((u64)bar_response_upper << 32) | bar_response;

				bar_size = ~(bar64 & PCI_BASE_ADDRESS_MEM_MASK) + 1;
				found_mem64 = 1;
			} else {
				bar_size = (u32)(~(bar_response & PCI_BASE_ADDRESS_MEM_MASK) + 1);
			}
			if (prefetch && (bar_response & PCI_BASE_ADDRESS_MEM_PREFETCH))
				bar_res = prefetch;
			else
				bar_res = mem;

			PRINTF("PCI Autoconfig: BAR %d, Mem, size=0x%llx, ", bar_nr, (u64)bar_size);

			if ((bar_size >= bar_size_lower) && (bar_size <= bar_size_upper)) {
				if (pciauto_region_allocate(bar_res, bar_size, &bar_value) == 0) {
					/* Write it out and update our limit */
					pci_hose_write_config_dword(hose, dev, bar, (u32)bar_value);
					PRINTF(" BAR written value=0x%8x, ", (u32)bar_value);

					if (found_mem64) {
						bar += 4;
#ifdef CONFIG_SYS_PCI_64BIT
						pci_hose_write_config_dword(hose, dev, bar, (u32)(bar_value>>32));
#else
						/*
						 * If we are a 64-bit decoder then increment to the
						 * upper 32 bits of the bar and force it to locate
						 * in the lower 4GB of memory.
						 */
						pci_hose_write_config_dword(hose, dev, bar, 0x00000000);
#endif
					}
				}
			}
		}

		PRINTF("\n");

		bar_nr++;
	}
}

void fix_pci_bars(void)
{
	struct pci_controller *hose = gd->ppc440_hose;
	unsigned int found_multi=0,problem=0,sec_func=0;
	unsigned short vendor, class;
	unsigned char header_type;
	pci_dev_t dev;
	u32 bar0;

	PRINTF("fix_pci_bars ++++++++++++++++++++++++++++++++++\n");

	for (dev =	PCI_BDF(0,0,0);
		 dev <	PCI_BDF(2,PCI_MAX_PCI_DEVICES-1,PCI_MAX_PCI_FUNCTIONS-1);
		 dev += PCI_BDF(0,0,1)) {

		if (pci_skip_dev(hose, dev))
			continue;

		if (PCI_FUNC(dev) && !found_multi)
			continue;

		pci_hose_read_config_byte(hose, dev, PCI_HEADER_TYPE, &header_type);

		pci_hose_read_config_word(hose, dev, PCI_VENDOR_ID, &vendor);

		if (vendor != 0xffff && vendor != 0x0000) {

			if (!PCI_FUNC(dev))
				found_multi = header_type & 0x80;

			pci_hose_read_config_word(hose, dev, PCI_CLASS_DEVICE, &class);

			PRINTF("PCI Scan: Found Bus %d, Device %d, Function %d - %x\n",
				PCI_BUS(dev), PCI_DEV(dev), PCI_FUNC(dev), class );

			if (((PCI_BUS(dev)==2) && (PCI_DEV(dev)==0) && (PCI_FUNC(dev)==0)) &&
				((class==0x300)	|| (class==0x380)))  {
				pci_read_config_dword(dev, PCI_BASE_ADDRESS_0, &bar0);
				bar0 = bar0 & 0xfffff000;
				PRINTF("BAR0 = %8x\n",bar0);
				problem +=1;
			}

			if (((PCI_BUS(dev)==2) && (PCI_DEV(dev)==0) && (PCI_FUNC(dev)==1)) &&
				((class==0x300)	|| (class==0x380))) sec_func =1;

			if ((PCI_BUS(dev)==0) && (PCI_DEV(dev)==0xc) && (PCI_FUNC(dev)==0)) {
				pci_read_config_dword(dev, PCI_BASE_ADDRESS_0, &bar0);
				bar0 = bar0 & 0xfffff000;
				PRINTF("BAR0 = %8x\n",bar0);

				if ((bar0 == 0) || (bar0 >= 0xA8000000)) problem +=1;
			}
		}
	}

	PRINTF("problem = %d\n",problem);

	if (problem >= 2)
	{
		pciauto_config_init(hose);

		/* setup MEM SPACE for PCI gfx card (big BARs) */
		dev =  PCI_BDF(2,0,0);
		pciauto_setup_device_mem(hose, dev, 6, hose->pci_mem, hose->pci_prefetch, hose->pci_io, 0x00100000, 0xFFFFFFFF);

		if (sec_func) {
			dev =  PCI_BDF(2,0,1);
			pciauto_setup_device_mem(hose, dev, 6, hose->pci_mem, hose->pci_prefetch, hose->pci_io, 0x00100000, 0xFFFFFFFF);
		}

		/* setup MEM SPACE for the onboard gfx card */
		dev =  PCI_BDF(0,0xc,0);
		pciauto_setup_device(hose, dev, 6, hose->pci_mem, hose->pci_prefetch, hose->pci_io);

		/* setup MEM SPACE for PCI gfx card (small BARs) */
		dev =  PCI_BDF(2,0,0);
		pciauto_setup_device_mem(hose, dev, 6, hose->pci_mem, hose->pci_prefetch, hose->pci_io, 0x0, 0x000FFFFF);

		if (sec_func) {
			dev =  PCI_BDF(2,0,1);
			pciauto_setup_device_mem(hose, dev, 6, hose->pci_mem, hose->pci_prefetch, hose->pci_io, 0x0, 0x000FFFFF);
		}
	}
}

void assign_pci_irq (void)
{
	u8  class;
	int BusNum;
	int Device;
	int Function;
	int reg_num = 0;
	int index = 0;
	unsigned char HeaderType;
	unsigned short VendorID;
	pci_dev_t dev;
	int irq_array[4] = { 50, 27, 29, 50 };

	reg_num = gd->ppc440_hose->region_count;

	pci_set_region(gd->ppc440_hose->regions + reg_num++,
			0x00000000,
			CFG_ISA_IO_BASE_ADDRESS,
			0x01000000,
			PCI_REGION_IO);

	gd->ppc440_hose->region_count = reg_num;

	// On Board fixed PCI devices ----------------

	// Primary PCI 66 Mhz bus --------------------

	// Bridge
	if ((dev = pci_find_device(0x12d8,0x8150,0)) >= 0)
	{
		pci_write_config_byte(dev, PCI_LATENCY_TIMER, 0xff);
		pci_write_config_byte(dev, PCI_SEC_LATENCY_TIMER, 0xff);
		pci_write_config_byte(dev, PCI_CACHE_LINE_SIZE, 0x10);

		// shut down unused secondary clock sources
		// clocks number 0,1,2,3 and 9 are used on
		// the Sam440ep board

		pci_write_config_word(dev, 0x68, 0x1f00);
	}

	// ATI Radeon M9
	if ((dev = pci_find_device(0x1002,0x4c66,0)) >= 0)
	{
		pci_write_config_byte(dev, PCI_INTERRUPT_LINE, 25);
		pci_write_config_byte(dev, PCI_LATENCY_TIMER, 0x80);
		pci_write_config_byte(dev, PCI_CACHE_LINE_SIZE, 0x10);
	}

	// Silicon Image 3114
	if ((dev = pci_find_device(0x1095,0x3114,0)) >= 0)
	{
		pci_write_config_byte(dev, PCI_INTERRUPT_LINE, 26);
		pci_write_config_byte(dev, PCI_LATENCY_TIMER, 0x40);
	}

	// Secondary PCI 33 Mhz bus ------------------

	// Cirrus Logic Audio on PCI33 Secondary Bus
	if ((dev = pci_find_device(0x1013,0x6005,0)) >= 0)
	{
		pci_write_config_byte(dev, PCI_INTERRUPT_LINE, 27);
		pci_write_config_byte(dev, PCI_LATENCY_TIMER, 0x20); // 0x18
	}

	// USB ISP1561 on PCI33 Secondary Bus
	if ((dev = pci_find_device(0x1131,0x1561,0)) >= 0)
	{
		pci_write_config_byte(PCI_BDF(0x01,0x05,0x00), PCI_INTERRUPT_LINE, 29);
		pci_write_config_byte(PCI_BDF(0x01,0x05,0x01), PCI_INTERRUPT_LINE, 29);
		pci_write_config_byte(PCI_BDF(0x01,0x05,0x02), PCI_INTERRUPT_LINE, 29);

		pci_write_config_byte(PCI_BDF(0x01,0x05,0x00), PCI_LATENCY_TIMER, 0x20); // 0x2A
		pci_write_config_byte(PCI_BDF(0x01,0x05,0x01), PCI_LATENCY_TIMER, 0x20); // 0x2a
		pci_write_config_byte(PCI_BDF(0x01,0x05,0x02), PCI_LATENCY_TIMER, 0x20); // 0x10

		// IMPORTANT FIX for a problem with the ISP1561 USB chip
		// refer to Errata 11

		pci_write_config_byte(PCI_BDF(0x01,0x05,0x00), 0x41, 0x00);
	}

	// Optional PCI devices on Expansion Slot ----

	for (BusNum = 1; BusNum < 4; BusNum++)
	{
		for (Device = 0; Device < PCI_MAX_PCI_DEVICES; Device++)
		{
			HeaderType = 0;

			for (Function = 0; Function < PCI_MAX_PCI_FUNCTIONS; Function++)
			{
				if (Function && !(HeaderType & 0x80))
					break;

				dev = PCI_BDF(BusNum, Device, Function);

				if (dev)
				{
					pci_read_config_word(dev, PCI_VENDOR_ID, &VendorID);
					if ((VendorID == 0xFFFF) || (VendorID == 0x0000))
						continue;

					if (!Function)
						pci_read_config_byte(dev, PCI_HEADER_TYPE, &HeaderType);

					pci_read_config_byte(dev, PCI_CLASS_CODE, &class);

					if (class != PCI_BASE_CLASS_BRIDGE)
					{
						if ((BusNum == 1) && ((Device == 4) || (Device == 5))) continue;

						pci_write_config_byte(dev, PCI_INTERRUPT_LINE, irq_array[index++]);
						index &= 3;

					}
					else if ((BusNum == 1) && (Device == 6))
					{
					    pci_write_config_byte(dev, PCI_INTERRUPT_LINE, irq_array[index]);
					}

					pci_write_config_byte(dev, PCI_LATENCY_TIMER, 0x20);
				}
			}
		}
	}

	// special configuration for PCI-PCI Express bridge PEX8112
	if ((dev = pci_find_device(0x10b5,0x8112,0)) >= 0)
	{
		pci_write_config_byte(dev, PCI_CACHE_LINE_SIZE, 0x10);
		pci_write_config_byte(dev, PCI_LATENCY_TIMER, 0xf0);
		pci_write_config_byte(dev, PCI_SEC_LATENCY_TIMER, 0xf8);
		pci_write_config_byte(dev, 0x48, 0x13);
		pci_write_config_byte(dev, 0x84, 0x0c); // index = 0x100c
		pci_write_config_dword(dev, 0x88, 0xcf008020); // data
		
		struct pci_controller *hose = gd->ppc440_hose;
		pciauto_setup_device_mem(hose, dev, 2, hose->pci_mem, NULL, NULL, 0x00010000, 0xFFFFFFFF);
	}
}

int last_stage_init (void)
{
	uchar buf;
	int ret;

	u32 val = 0x00010000;
	mtspr(mmucr,val);

	// fix possible menuboot_cmd misconfiguration
	char *s = getenv("menuboot_cmd");
	if ((!s) ||
	    ((s) && (strlen(s) < 3)) ||
	    ((s) && (strcmp(s,"noboot") == 0)))

	   setenv("menuboot_cmd","boota");

	// fix possible ext_gfx_irq misconfiguration
	s = getenv("ext_gfx_irq");
	if (!s) setenv("ext_gfx_irq","enabled");

	//fix_pci_bars();
	assign_pci_irq();

	// cache on -------------------------------------------
	change_tlb(0, 256*1024*1024, 0);

	ret = init_vga(gd->ppc440_hose);

	if (ret != -1)
	{
		unsigned short cmd = 0;
		pci_dev_t dev;

		drv_video_init();

	    // if there is a SM502 gfc card, switch it off
	    // since booting with the SM502 active isn't supported yet
		if ((dev = pci_find_device(PCI_VENDOR_SM, PCI_DEVICE_SM501, 0)) >= 0)
		{
			pci_read_config_word(dev, PCI_COMMAND, &cmd);
			cmd &= ~(PCI_COMMAND_IO|PCI_COMMAND_MEMORY);
			pci_write_config_word(dev, PCI_COMMAND, cmd);
		}
	}

	// custom silent mode ---------------------------------
	int hush = 0;
	s = getenv("hush");
	if (s) hush = atoi(s);
	if (hush) {
	    s = getenv("stdout");
	    if ((s) && (strncmp(s,"vga",3) == 0))
	        gd->flags |= GD_FLG_SILENT;
	 }

#ifdef SAMLOGO
	// Welcome Screen -------------------------------------
	if (fbi)
	{
		unsigned int xx, yy, xoff = 0, yoff = 0;

		if (gd->flags & GD_FLG_SILENT) {
		    xoff = (fbi->XSize-LOGO_WIDTH) / 2;
		    yoff = (fbi->YSize-LOGO_HEIGHT) / 9;
		}
		else puts("\n\n\n\n");

		for (xx = 0; xx < LOGO_WIDTH; xx++)
		{
			for (yy = 0; yy < LOGO_HEIGHT; yy++)
			{
				buf = 255 - logo_acube[xx + (LOGO_HEIGHT-yy-1)*LOGO_WIDTH];
				*((char *)(fbi->BaseAddress + (xx+xoff) + (yy+yoff)*fbi->XSize*(fbi->BitsPerPixel/8))) = buf;
			}
		}
	}
#endif

	// cache off ------------------------------------------
	change_tlb(0, 256*1024*1024, TLB_WORD2_I_ENABLE);

	// USB keyboard ---------------------------------------
	if (gd->flags & GD_FLG_SILENT) {
		gd->flags &= ~GD_FLG_SILENT;
		console_row = 29;
		console_col = 28;
		puts("Init USB... ");
		gd->flags |= GD_FLG_SILENT;
	}

	ret = usb_init();

#ifdef CONFIG_USB_STORAGE
	// try to recognize storage devices immediately
	if (ret >= 0)
	{
		char *s = getenv("scan_usb_storage");
		if (s) usb_stor_scan(1);
	}
#endif

	// prints last switch off and restart the TimeKeeper registers on RTC
	puts("\nLast Switch off: ");
	do_date(NULL, 0, 1, NULL);

	i2c_read(0x68, 0x0C, 1, &buf, 1);
	buf = buf & 0xbf; // clear HT bit in register 0x0c
	i2c_write(0x68, 0x0c, 1, &buf, 1);

	puts("\n");

	// Init SATA controller ----------------------
	if (gd->flags & GD_FLG_SILENT) {
		gd->flags &= ~GD_FLG_SILENT;
		puts("Done - Init SATA... ");
		gd->flags |= GD_FLG_SILENT;
	}

	init_ATA_controllers();

	if (gd->flags & GD_FLG_SILENT) {
		gd->flags &= ~GD_FLG_SILENT;
		puts("Done\n");
		gd->flags |= GD_FLG_SILENT;
	}

	return 0;
}

void board_reset(void)
{
	int ii;
	u16 value;
	pci_dev_t dev;
	unsigned long addr = 0xfffffffc;

	if ((dev = pci_find_device(0x1014,0x027f,0)) >= 0)
	{
		pci_read_config_word(dev, 0x60, &value);
		pci_write_config_word(dev, 0x60, value | (1<<12));

		for (ii = 0; ii < 100; ii++)
			__asm volatile ("sync");

		pci_write_config_word(dev, 0x60, value & ~(1<<12));
	}

	__asm volatile ("mtspr    26, %0"         :: "r" (addr));
	__asm volatile ("li       4, (1 << 6)"    ::: "r4");
	__asm volatile ("mtspr    27, 4");
	__asm volatile ("rfi");

	while(1);       /* not reached */
}
