/*
 * (C) Copyright 2008-2009
 * Max Tretene, ACube Systems Srl. mtretene@acube-systems.com.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <ppc4xx.h>
#include <asm/processor.h>
#include <asm/mmu.h>
#include <i2c.h>
#include <part.h>
#include <command.h>
#include <usb.h>
#include <video_fb.h>
#include "spd_sdram.h"
#include "Sam440ep_flex.h"
#include "../common/vesa.h"

#define SAMLOGO
#ifdef SAMLOGO
#include "../common/logo_acube.h"
#else
#define LOGO_WIDTH 176
#define LOGO_HEIGHT 48
unsigned char logo_acube[LOGO_WIDTH * LOGO_HEIGHT] = { 0 };
#endif

DECLARE_GLOBAL_DATA_PTR;

#undef SAM_DEBUG

#ifndef CFG_NO_FLASH
extern flash_info_t flash_info[CFG_MAX_FLASH_BANKS]; /* info for FLASH chips	*/
#endif

extern int console_col; /* cursor col */
extern int console_row; /* cursor row */

extern u32 *fb_base_phys_sm502;
extern unsigned char SM502INIT;
extern pci_dev_t dev_sm502;

unsigned char SM502 = 0;
struct FrameBufferInfo *fbi = NULL;

extern pci_dev_t pci_find_bridge_for_bus(struct pci_controller *hose, int busnr);

int board_early_init_f(void)
{
	register uint reg;

	/*--------------------------------------------------------------------
	 * Setup the external bus controller/chip selects
	 *-------------------------------------------------------------------*/

	mtdcr(ebccfga, xbcfg);
	reg = mfdcr(ebccfgd);
	mtdcr(ebccfgd, reg | 0x04000000);	// Set ATC

	mtebc(pb0ap, 0x06017240);   // 0x03017240);	// CS0 - EEProm
	mtebc(pb0cr, 0xfff18000);	// BAS=0xfff 1MB ro 8-bit

	mtebc(pb2ap, 0x00000000);	// CS2 - SmartMedia
	mtebc(pb2cr, 0x00000000);	//

	mtebc(pb3ap, 0x01057240);	// CS3 - FPGA
	mtebc(pb3cr, 0xfe05a000);	// BAS=0xfe000000 4MB r/w 16-bit

	mtebc(pb4ap, 0x00000000);
	mtebc(pb4cr, 0x00000000);

	mtebc(pb5ap, 0x00000000);
	mtebc(pb5cr, 0x00000000);

	/*--------------------------------------------------------------------
	 * Setup the GPIO pins
	 *-------------------------------------------------------------------*/

	/*CPLD cs */
	/*setup Address lines for flash size 64Meg. */
	out32(GPIO0_OSRL, in32(GPIO0_OSRL) | 0x50010000);
	out32(GPIO0_TSRL, in32(GPIO0_TSRL) | 0x50010000);
	out32(GPIO0_ISR1L, in32(GPIO0_ISR1L) | 0x50000000);

	/*setup emac */
	out32(GPIO0_TCR, in32(GPIO0_TCR) | 0xC080);
	out32(GPIO0_TSRL, in32(GPIO0_TSRL) | 0x40);
	out32(GPIO0_ISR1L, in32(GPIO0_ISR1L) | 0x55);
	out32(GPIO0_OSRH, in32(GPIO0_OSRH) | 0x50004000);
	out32(GPIO0_ISR1H, in32(GPIO0_ISR1H) | 0x00440000);

	/*UART1 */
	out32(GPIO1_TCR, in32(GPIO1_TCR) | 0x02000000);
	out32(GPIO1_OSRL, in32(GPIO1_OSRL) | 0x00080000);
	out32(GPIO1_ISR2L, in32(GPIO1_ISR2L) | 0x00010000);

	/* external interrupts IRQ0...4 and IRQ6...7 */
	out32(GPIO1_TCR, in32(GPIO1_TCR) & ~0x00fe0000);
	out32(GPIO1_TSRL, in32(GPIO1_TSRL) & ~0x0000fffc);
	out32(GPIO1_ISR1L, in32(GPIO1_ISR1L) | 0x00005554);

	/*--------------------------------------------------------------------
	 * Setup the interrupt controller polarities, triggers, etc.
	 *-------------------------------------------------------------------*/
	mtdcr(uic0sr, 0xffffffff);	/* clear all */
	mtdcr(uic0er, 0x00000000);	/* disable all */
	mtdcr(uic0cr, 0x00000009);  /* ATI & UIC1 crit are critical */
	mtdcr(uic0pr, 0xffbffe03);
	mtdcr(uic0tr, 0x00000008);	/* per ref-board manual */
	mtdcr(uic0vr, 0x00000001);	/* int31 highest, base=0x000 */
	mtdcr(uic0sr, 0xffffffff);	/* clear all */

	mtdcr(uic1sr, 0xffffffff);	/* clear all */
	mtdcr(uic1er, 0x00000000);	/* disable all */
	mtdcr(uic1cr, 0x00000000);	/* all non-critical */
	mtdcr(uic1pr, 0xffffe6af);
	mtdcr(uic1tr, 0x06000140);	/* per ref-board manual */
	mtdcr(uic1vr, 0x00000001);	/* int31 highest, base=0x000 */
	mtdcr(uic1sr, 0xffffffff);	/* clear all */

	/*--------------------------------------------------------------------
	 * Setup other serial configuration
	 *-------------------------------------------------------------------*/
	mfsdr(sdr_pci0, reg);
	mtsdr(sdr_pci0, 0x80000000 | reg);	/* PCI arbiter enabled */
	mtsdr(sdr_pfc0, 0x00003e00);	/* Pin function */
	mtsdr(sdr_pfc1, 0x00088000);	/* Pin function: UART0 has 8 pins */

	return 0;
}

int misc_init_r (void)
{
	return 0;
}

int checkboard(void)
{
	sys_info_t sysinfo;

	get_sys_info(&sysinfo);
	puts ("Board: Sam440ep-flex\n");

	return (0);
}

/*************************************************************************
 *  long int initdram
 *
 ************************************************************************/
long int initdram(int board)
{
	long dram_size = 0;

	unsigned char cfg_simulate_spd_eeprom[32] = {
		0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00 };

	dram_size = spd_sdram(cfg_simulate_spd_eeprom);

	return dram_size;
}

#if defined(CFG_DRAM_TEST)
int testdram(void)
{
	unsigned long *mem = (unsigned long *)0;
	const unsigned long kend = (1024 / sizeof(unsigned long));
	unsigned long k, n;
	unsigned long value;
	int error = 0;

	char *p = getenv("testram");
	if (p != NULL) {
		mtmsr(0);

		for (k = 0; k < gd->ram_size/1024; //CFG_KBYTES_SDRAM * CFG_SDRAM_BANKS;
		     ++k, mem += (1024 / sizeof(unsigned long))) {
			if ((k & 1023) == 0) {
				printf("Testing: %3d MB\r", k / 1024);
			}

			for (n = 0; n < kend; ++n) {
				value = (n << 24 | n << 16 | n << 8 | n);
				mem[n] = value;
			}

			for (n = 0; n < kend; ++n) {
				value = (n << 24 | n << 16 | n << 8 | n);
				if (mem[n] != value) {
					printf("SDRAM test fails at %08x, wrote %08x read %08x\n",
					       (uint) & mem[n],value,mem[n]);
					error = 1;
				}
			}
		}

		if (error == 0) puts ("SDRAM test passes\n");
		return error;
	}
	else return 0;
}
#endif

/*************************************************************************
 *  pci_pre_init
 *
 *  This routine is called just prior to registering the hose and gives
 *  the board the opportunity to check things. Returning a value of zero
 *  indicates that things are bad & PCI initialization should be aborted.
 *
 *	Different boards may wish to customize the pci controller structure
 *	(add regions, override default access routines, etc) or perform
 *	certain pre-initialization actions.
 *
 ************************************************************************/
#if defined(CONFIG_PCI) && defined(CFG_PCI_PRE_INIT)
int pci_pre_init(struct pci_controller *hose)
{
	unsigned long addr;

	/*-------------------------------------------------------------------------+
	  | Set priority for all PLB3 devices to 0.
	  | Set PLB3 arbiter to fair mode.
	  +-------------------------------------------------------------------------*/
	mfsdr(sdr_amp1, addr);
	mtsdr(sdr_amp1, (addr & 0x000000FF) | 0x0000FF00);
	addr = mfdcr(plb3_acr);
	mtdcr(plb3_acr, addr | 0x88000000); // high load enable - was 0x80000000

	/*-------------------------------------------------------------------------+
	  | Set priority for all PLB4 devices to 0.
	  +-------------------------------------------------------------------------*/
	mfsdr(sdr_amp0, addr);
	mtsdr(sdr_amp0, (addr & 0x000000FF) | 0x0000FF00);
	addr = mfdcr(plb4_acr) | 0xa0000000;	/* Was 0x8---- */
	mtdcr(plb4_acr, addr);

	/*-------------------------------------------------------------------------+
	  | Set Nebula PLB4 arbiter to fair mode.
	  +-------------------------------------------------------------------------*/
	/* Segment0 */
	addr = (mfdcr(plb0_acr) & ~plb0_acr_ppm_mask) | plb0_acr_ppm_fair;
	addr = (addr & ~plb0_acr_hbu_mask) | plb0_acr_hbu_enabled;
	addr = (addr & ~plb0_acr_rdp_mask) | plb0_acr_rdp_4deep;
	addr = (addr & ~plb0_acr_wrp_mask) | plb0_acr_wrp_2deep;
	mtdcr(plb0_acr, addr);

	/* Segment1 */
	addr = (mfdcr(plb1_acr) & ~plb1_acr_ppm_mask) | plb1_acr_ppm_fair;
	addr = (addr & ~plb1_acr_hbu_mask) | plb1_acr_hbu_enabled;
	addr = (addr & ~plb1_acr_rdp_mask) | plb1_acr_rdp_4deep;
	addr = (addr & ~plb1_acr_wrp_mask) | plb1_acr_wrp_2deep;
	mtdcr(plb1_acr, addr);

	return 1;
}
#endif				/* defined(CONFIG_PCI) && defined(CFG_PCI_PRE_INIT) */

/*************************************************************************
 *  pci_target_init
 *
 *	The bootstrap configuration provides default settings for the pci
 *	inbound map (PIM). But the bootstrap config choices are limited and
 *	may not be sufficient for a given board.
 *
 ************************************************************************/
#if defined(CONFIG_PCI) && defined(CFG_PCI_TARGET_INIT)
void pci_target_init(struct pci_controller *hose)
{
	/*--------------------------------------------------------------------------+
	 * Set up Direct MMIO registers
	 *--------------------------------------------------------------------------*/
	/*--------------------------------------------------------------------------+
	  | PowerPC440 EP PCI Master configuration.
	  | Map one 1Gig range of PLB/processor addresses to PCI memory space.
	  |   PLB address 0xA0000000-0xDFFFFFFF ==> PCI address 0xA0000000-0xDFFFFFFF
	  |   Use byte reversed out routines to handle endianess.
	  | Make this region non-prefetchable.
	  +--------------------------------------------------------------------------*/
	out32r(PCIX0_PMM0MA,    0x00000000);       /* PMM0 Mask/Attribute - disabled b4 setting */
	out32r(PCIX0_PMM0LA,    CFG_PCI_MEMBASE);  /* PMM0 Local Address */
	out32r(PCIX0_PMM0PCILA, CFG_PCI_MEMBASE);  /* PMM0 PCI Low Address */
	out32r(PCIX0_PMM0PCIHA, 0x00000000);       /* PMM0 PCI High Address */
	out32r(PCIX0_PMM0MA,    0xE0000001);       /* 512M + No prefetching, and enable region */

	out32r(PCIX0_PMM1MA,    0x00000000);       /* PMM1 Mask/Attribute - disabled b4 setting */
	out32r(PCIX0_PMM1LA,    CFG_PCI_MEMBASE2); /* PMM1 Local Address */
	out32r(PCIX0_PMM1PCILA, CFG_PCI_MEMBASE2); /* PMM1 PCI Low Address */
	out32r(PCIX0_PMM1PCIHA, 0x00000000);       /* PMM1 PCI High Address */
	out32r(PCIX0_PMM1MA,    0xE0000001);       /* 512M + No prefetching, and enable region */

	out32r(PCIX0_PMM2MA,    0);                /* PMM2 Mask/Attribute - disabled */
	out32r(PCIX0_PMM2LA,    0);                /* PMM2 Local Address */
	out32r(PCIX0_PMM2PCILA, 0);                /* PMM2 PCI Low Address */
	out32r(PCIX0_PMM2PCIHA, 0);                /* PMM2 PCI High Address */

	out32r(PCIX0_PTM1MS, 0x00000001);          /* Memory Size/Attribute */
	out32r(PCIX0_PTM1LA, 0);                   /* Local Addr. Reg */
	out32r(PCIX0_PTM2MS, 0);                   /* Memory Size/Attribute */
	out32r(PCIX0_PTM2LA, 0);                   /* Local Addr. Reg */

	/*--------------------------------------------------------------------------+
	 * Set up Configuration registers
	 *--------------------------------------------------------------------------*/

	/* Program the board's subsystem id/vendor id */
	pci_write_config_word(0, PCI_SUBSYSTEM_VENDOR_ID,
			      CFG_PCI_SUBSYS_VENDORID);
	pci_write_config_word(0, PCI_SUBSYSTEM_ID, CFG_PCI_SUBSYS_ID);

	/* Configure command register as bus master */
	pci_write_config_word(0, PCI_COMMAND, PCI_COMMAND_MASTER);

	/* 240nS PCI clock */
	pci_write_config_byte(0, PCI_LATENCY_TIMER, 0x0);

	/* No error reporting */
	pci_write_config_word(0, PCI_ERREN, 0);

	pci_write_config_dword(0, PCI_BRDGOPT2, 0x00000101);

}
#endif				/* defined(CONFIG_PCI) && defined(CFG_PCI_TARGET_INIT) */

/*************************************************************************
 *  pci_master_init
 *
 ************************************************************************/
#if defined(CONFIG_PCI) && defined(CFG_PCI_MASTER_INIT)
void pci_master_init(struct pci_controller *hose)
{
	unsigned short temp_short;

	/*--------------------------------------------------------------------------+
	  | Write the PowerPC440 EP PCI Configuration regs.
	  |   Enable PowerPC440 EP to be a master on the PCI bus (PMM).
	  |   Enable PowerPC440 EP to act as a PCI memory target (PTM).
	  +--------------------------------------------------------------------------*/
	pci_read_config_word(0, PCI_COMMAND, &temp_short);
	pci_write_config_word(0, PCI_COMMAND,
			      temp_short | PCI_COMMAND_MASTER |
			      PCI_COMMAND_MEMORY);
}
#endif				/* defined(CONFIG_PCI) && defined(CFG_PCI_MASTER_INIT) */

/*************************************************************************
 *  is_pci_host
 *
 *	This routine is called to determine if a pci scan should be
 *	performed. With various hardware environments (especially cPCI and
 *	PPMC) it's insufficient to depend on the state of the arbiter enable
 *	bit in the strap register, or generic host/adapter assumptions.
 *
 *	Rather than hard-code a bad assumption in the general 440 code, the
 *	440 pci code requires the board to decide at runtime.
 *
 *	Return 0 for adapter mode, non-zero for host (monarch) mode.
 *
 *
 ************************************************************************/
#if defined(CONFIG_PCI)
int is_pci_host(struct pci_controller *hose)
{
	/* Sam440ep is always configured as host. */
	return (1);
}
#endif				/* defined(CONFIG_PCI) */

/*************************************************************************
 *  hw_watchdog_reset
 *
 *	This routine is called to reset (keep alive) the watchdog timer
 *
 ************************************************************************/
#if defined(CONFIG_HW_WATCHDOG)
void hw_watchdog_reset(void)
{

}
#endif

void assign_pci_irq (void)
{
	u8 class, pin, bridge_secbus;
	int BusNum;
	int Device, Bridge;
	int Function;
	int reg_num = 0;
	int index = 0;
	unsigned char HeaderType;
	unsigned short VendorID;
	pci_dev_t dev, tmpdev;

	int irq_array[16][4] = {
		{ 25, 25, 25, 25 },  // .0
		{  0,  0,  0,  0 },  // .1
		{  0,  0,  0,  0 },  // .2
		{  0,  0,  0,  0 },  // .3
		{  0,  0,  0,  0 },  // .4
		{  0,  0,  0,  0 },  // .5
		{ 50, 27, 29, 50 },  // .6
		{ 29, 50, 29, 50 },  // .7
		{ 29, 50, 27, 50 },	 // .8
		{  0,  0,  0,  0 },  // .9
		{  0,  0,  0,  0 },  // .a
		{  0,  0,  0,  0 },  // .b
		{ 25, 25, 25, 25 },  // .c
		{  0,  0,  0,  0 },  // .d
		{  0,  0,  0,  0 },  // .e
		{  0,  0,  0,  0 },  // .f
	};

	reg_num = gd->ppc440_hose->region_count;

	pci_set_region(gd->ppc440_hose->regions + reg_num++,
			0x00000000,
			CFG_ISA_IO_BASE_ADDRESS,
			0x01000000,
			PCI_REGION_IO);

	gd->ppc440_hose->region_count = reg_num;

	// Primary PCI 66 Mhz bus -----------------------------

	// Bridge
	if ((dev = pci_find_device(0x12d8,0x8150,0)) >= 0)
	{
		pci_write_config_byte(dev, PCI_LATENCY_TIMER, 0xff);
		pci_write_config_byte(dev, PCI_SEC_LATENCY_TIMER, 0xff);
		pci_write_config_byte(dev, PCI_CACHE_LINE_SIZE, 0x04);

		// shut down unused secondary clock sources
		// clocks number 0,1,2,3, 4 and 9 are used on
		// the Sam440ep-flex board

		pci_write_config_word(dev, 0x68, 0x1e00);
	}

	// Silicon Image 3114
	if ((dev = pci_find_device(0x1095,0x3114,0)) >= 0)
	{
		pci_write_config_byte(dev, PCI_INTERRUPT_LINE, 26);
		pci_write_config_byte(dev, PCI_LATENCY_TIMER, 0x40);
	}

	// Secondary PCI 33 Mhz bus ---------------------------

	// Cirrus Logic Audio on PCI33 Secondary Bus
	if ((dev = pci_find_device(0x1013,0x6005,0)) >= 0)
	{
		pci_write_config_byte(dev, PCI_INTERRUPT_LINE, 27);
		pci_write_config_byte(dev, PCI_LATENCY_TIMER, 0x20);
	}

	// USB NEC on PCI33 Secondary Bus
	if ((dev = pci_find_device(0x1033,0x0035,0)) >= 0)
	{
		pci_write_config_byte(PCI_BDF(0x01,0x05,0x00), PCI_INTERRUPT_LINE, 29);
		pci_write_config_byte(PCI_BDF(0x01,0x05,0x01), PCI_INTERRUPT_LINE, 29);
		pci_write_config_byte(PCI_BDF(0x01,0x05,0x02), PCI_INTERRUPT_LINE, 29);

		pci_write_config_byte(PCI_BDF(0x01,0x05,0x00), PCI_LATENCY_TIMER, 0x20);
		pci_write_config_byte(PCI_BDF(0x01,0x05,0x01), PCI_LATENCY_TIMER, 0x20);
		pci_write_config_byte(PCI_BDF(0x01,0x05,0x02), PCI_LATENCY_TIMER, 0x20);

		pci_write_config_byte(PCI_BDF(0x01,0x05,0x00), 0xE0, 0x04);	// 4 USB ports setup
		pci_write_config_byte(PCI_BDF(0x01,0x05,0x00), 0xE4, 0x20);	// 48 Mhz clock
	}

	// Optional PCI devices on PCI Slots 33/66 Mhz --------

	for (BusNum = 0; BusNum < 5; BusNum++)
	{
		for (Device = 0; Device < PCI_MAX_PCI_DEVICES; Device++)
		{
			HeaderType = 0;

			for (Function = 0; Function < PCI_MAX_PCI_FUNCTIONS; Function++)
			{
				if (Function && !(HeaderType & 0x80))
					break;

				dev = PCI_BDF(BusNum, Device, Function);

				if (dev)
				{
					pci_read_config_word(dev, PCI_VENDOR_ID, &VendorID);
					if ((VendorID == 0xFFFF) || (VendorID == 0x0000))
						continue;

					if (!Function)
						pci_read_config_byte(dev, PCI_HEADER_TYPE, &HeaderType);

					pci_read_config_byte(dev, PCI_CLASS_CODE, &class);

					//if (class != PCI_BASE_CLASS_BRIDGE)
					{
						if ((BusNum == 0) && ((Device == 0x0a) || (Device == 0x0e))) continue;
						if ((BusNum == 1) && ((Device == 0x04) || (Device == 0x05))) continue;

						pci_read_config_byte(dev, PCI_INTERRUPT_PIN, &pin);

						if (pin > 0)
						{
						    tmpdev = pci_find_bridge_for_bus(gd->ppc440_hose,BusNum);

							if (tmpdev != PCI_ANY_ID) Bridge = PCI_DEV(tmpdev);
							else Bridge = 0; // assuming card on primary PCI bus

							if (Bridge == 0x0a) Bridge = Device;

							pci_write_config_byte(dev, PCI_INTERRUPT_LINE, irq_array[Bridge][pin-1]);
						}

						pci_write_config_byte(dev, PCI_LATENCY_TIMER, 0x80);
						pci_write_config_byte(dev, PCI_CACHE_LINE_SIZE, 0x20);
					}
				}
			}
		}
	}
}

void config_pex8112(void)
{
	pci_dev_t dev;

	// special configuration for PCI-PCI Express bridge PEX8112
	if ((dev = pci_find_device(0x10b5,0x8112,0)) >= 0)
	{
		pci_write_config_byte(dev, PCI_CACHE_LINE_SIZE, 0x10);
		pci_write_config_byte(dev, PCI_LATENCY_TIMER, 0xff);
		pci_write_config_byte(dev, PCI_SEC_LATENCY_TIMER, 0xff);
		pci_write_config_byte(dev, 0x48, 0x11);
		pci_write_config_byte(dev, 0x84, 0x0c); // index = 0x100c
		pci_write_config_dword(dev, 0x88, 0xcf008020); // data
	}
}

int last_stage_init (void)
{
	uchar buf;
	int jj, ret;

	u32 val = 0x00010000;
	mtspr(mmucr,val);

	// fix possible menuboot_cmd misconfiguration ---------
	char *s = getenv("menuboot_cmd");
	if ((!s) ||
	    ((s) && (strlen(s) < 3)) ||
	    ((s) && (strcmp(s,"noboot") == 0)))

	   setenv("menuboot_cmd","boota");

	// fix possible ext_gfx_irq misconfiguration
	s = getenv("ext_gfx_irq");
	if (!s) setenv("ext_gfx_irq","enabled");

	assign_pci_irq();

	// Graphic card(s) on PCI -----------------------------
	init_sm502();

	// cache on -------------------------------------------
	change_tlb(0, 256*1024*1024, 0);

	// x86 Graphic card on PCI ----------------------------
	ret = init_vga(gd->ppc440_hose);

	// active gfx card ------------------------------------
	SM502 = 0; // default PCI x86 card
	s = getenv("video_activate");
	if ((strcmp(s, "sm502") == 0) && (SM502INIT)) SM502 = 1;
	else if ((SM502INIT) && (ret == -1)) SM502 = 1;

	if (SM502 && SM502INIT)
	{
		fbi = (struct FrameBufferInfo *)(malloc(sizeof(struct FrameBufferInfo)));
	    if (fbi)
	    {
	       	fbi->BaseAddress   = fb_base_phys_sm502;
			fbi->XSize         = board_get_width();
			fbi->YSize         = board_get_height();
			fbi->BitsPerPixel  = 8;
			fbi->Modulo		   = board_get_width();

			drv_video_init();
		}
	}
	else
	{
	    if (SM502INIT)
	    {
	        // shutdown onboard gfx card
	        unsigned short cmd;

			pci_read_config_word(dev_sm502, PCI_COMMAND, &cmd);
			cmd &= ~(PCI_COMMAND_IO|PCI_COMMAND_MEMORY);
			pci_write_config_word(dev_sm502, PCI_COMMAND, cmd);
	    }

		if (ret >= 0)
		{
			drv_video_init();
		}
	}

	// custom silent mode ---------------------------------
	int hush = 0;
	s = getenv("hush");
	if (s) hush = atoi(s);
	if (hush) {
	    s = getenv("stdout");
	    if ((s) && (strncmp(s,"vga",3) == 0))
	        gd->flags |= GD_FLG_SILENT;
	 }

#ifdef SAMLOGO
	// Welcome Screen -------------------------------------
	if (fbi)
	{
		unsigned int xx, yy, xoff = 0, yoff = 0;

		if (gd->flags & GD_FLG_SILENT) {
		    xoff = (fbi->XSize-LOGO_WIDTH) / 2;
		    yoff = (fbi->YSize-LOGO_HEIGHT) / 9;
		}
		else puts("\n\n\n\n");

		for (xx = 0; xx < LOGO_WIDTH; xx++)
		{
			for (yy = 0; yy < LOGO_HEIGHT; yy++)
			{
				buf = 255 - logo_acube[xx + (LOGO_HEIGHT-yy-1)*LOGO_WIDTH];
				*((char *)(fbi->BaseAddress + (xx+xoff) + (yy+yoff)*fbi->XSize*(fbi->BitsPerPixel/8))) = buf;
			}
		}
	}
#endif

	// cache off ------------------------------------------
	change_tlb(0, 256*1024*1024, TLB_WORD2_I_ENABLE);

	// Catweasel keyboard ---------------------------------
	ret = catw_kb_init();

	// USB keyboard ---------------------------------------
	if (gd->flags & GD_FLG_SILENT) {
		gd->flags &= ~GD_FLG_SILENT;
		console_row = 29;
		console_col = 28;
		puts("Init USB... ");
		gd->flags |= GD_FLG_SILENT;
	}

	ret = usb_init();

#ifdef CONFIG_USB_STORAGE
	// try to recognize storage devices immediately
	if (ret >= 0)
	{
		s = getenv("scan_usb_storage");
		if (s) usb_stor_scan(1);
	}
#endif

	// prints last switch off and restart the TimeKeeper registers on RTC
	puts("\nLast Switch off: ");
	do_date(NULL, 0, 1, NULL);

	i2c_read(0x68, 0x0C, 1, &buf, 1);
	buf = buf & 0xbf; // clear HT bit in register 0x0c
	i2c_write(0x68, 0x0c, 1, &buf, 1);

	puts("\n");

	// Init SATA controller ----------------------
	if (gd->flags & GD_FLG_SILENT) {
		gd->flags &= ~GD_FLG_SILENT;
		puts("Done - Init SATA... ");
		gd->flags |= GD_FLG_SILENT;
	}

	init_ATA_controllers();

	if (gd->flags & GD_FLG_SILENT) {
		gd->flags &= ~GD_FLG_SILENT;
		puts("Done\n");
		gd->flags |= GD_FLG_SILENT;
	}

	config_pex8112();

	return 0;
}

void board_reset(void)
{
	int ii;
	u16 value;
	pci_dev_t dev;
	unsigned long addr = 0xfffffffc;

	if ((dev = pci_find_device(0x1014,0x027f,0)) >= 0)
	{
		pci_read_config_word(dev, 0x60, &value);
		pci_write_config_word(dev, 0x60, value | (1<<12));

		for (ii = 0; ii < 100; ii++)
			__asm volatile ("sync");

		pci_write_config_word(dev, 0x60, value & ~(1<<12));
	}

	__asm volatile ("mtspr    26, %0"         :: "r" (addr));
	__asm volatile ("li       4, (1 << 6)"    ::: "r4");
	__asm volatile ("mtspr    27, 4");
	__asm volatile ("rfi");

	while(1);       /* not reached */
}
