#define _STDDEF_H
#include <common.h>
#include "glue.h"
#include "scitech/include/x86emu/x86emu.h"
#include "x86interface.h"
/*
 * This isn't nice, but there are a lot of incompatibilities in the U-Boot and scitech include
 * files that this is the only really workable solution.
 * Might be cleaned out later.
 */

#undef DEBUG

//#define SINGLESTEP
//#define FORCE_SINGLESTEP

#undef IO_LOGGING
#undef MEM_LOGGING

#ifdef IO_LOGGING
#define LOGIO(port, format, args...) if (dolog(port)) _printf(format , ## args)
#else
#define LOGIO(port, format, args...)
#endif

#ifdef MEM_LOGGIN
#define LOGMEM(format, args...) _printf(format , ## args)
#else
#define LOGMEM(format, args...)
#endif

#define log_printf(format, args...)  if (getenv("x86_log")) printf(format, ## args);

#ifdef DEBUG
#warning "Debuggin enabled in " __FILE__
#define PRINTF(format, args...) _printf(format , ## args)
#else
#define PRINTF(format, argc...)
#endif

typedef unsigned char UBYTE;
typedef unsigned short UWORD;
typedef unsigned long ULONG;

typedef char BYTE;
typedef short WORT;
typedef long LONG;

#define EMULATOR_MEM_SIZE       (1024*1024)
#define EMULATOR_BIOS_OFFSET    0xC0000
#define EMULATOR_STRAP_OFFSET   0x30000
#define EMULATOR_STACK_OFFSET   0x20000
#define EMULATOR_LOGO_OFFSET    0x40000 // If you change this, change the strap code, too

extern int tstc(void);
extern int getc(void);
extern unsigned char video_get_attr(void);
extern void find_radeon_values(pci_dev_t dev, u8 * rom_addr);
extern void reloc_mode_table(void *reloc_addr);

extern int pci_card;

#include "x86interface.h"

extern void bios_set_mode(int mode);

void sam440_remove_init_data(void)
{

}

void setup_tlb_for_cache(int enable)
{
	// not used
}

//Forward declaration
void do_inout(void);

int abs(int x)
{
	if (x < 0)
	   return -x;

	return x;
}

void cons_gets(char *buffer)
{
	int i = 0;
	char c = 0;

	buffer[0] = 0;
	if (getenv("x86_runthru")) return; //FIXME:
	while (c != 0x0D && c != 0x0A)
	{
		while (!tstc());
		c = getc();
		if (c>=32 && c < 127)
		{
			buffer[i] = c;
			i++;
			buffer[i] = 0;
			putc(c);
		}
		else
		{
			if (c == 0x08)
			{
			if (i>0) i--;
			buffer[i] = 0;
			}
		}
	}
	buffer[i] = '\n';
	buffer[i+1] = 0;
}

char *bios_date = "08/14/02";
UBYTE model = 0xFC;
UBYTE submodel = 0x00;

static int log_init = 0;
static int log_do = 0;
static int log_low = 0;

int dolog(int port)
{
	if (log_init && log_do)
	{
		if (log_low && port > 0x400) return 0;
		return 1;
	}

	if (!log_init)
	{
		log_init = 1;
		log_do = (getenv("x86_logio") != (char *)0);
		log_low = (getenv("x86_loglow") != (char *)0);
		if (log_do)
		{
			if (log_low && port > 0x400) return 0;
			return 1;
		}
	}
	return 0;
}

static u32 dummy;

u32 screen_addr(u32 addr)
{
	return &dummy;
}

// Converts an emulator address to a physical address.
// Handles all special cases (bios date, model etc), and might need work
u32 memaddr(u32 addr)
{
	if (addr >= 0xA0000 && addr < 0xC0000)
	return screen_addr(addr); //CFG_ISA_IO_BASE_ADDRESS + addr;
	else if (addr >= 0xFFFF5 && addr < 0xFFFFE)
	{
		return (u32)bios_date+addr-0xFFFF5;
	}
	else if (addr == 0xFFFFE)
	return (u32)&model;
	else if (addr == 0xFFFFF)
	return (u32)&submodel;
	else if (addr >= 0x80000000)
	{
		//printf("Warning: High memory access at 0x%x\n", addr);
		return addr;
	}
	else
	return (u32)M.mem_base+addr;
}

u8 A1_rdb(u32 addr)
{
	u8 a = in8((UBYTE *)memaddr(addr));
	LOGMEM("rdb: %x -> %x\n", addr, a);
	return a;
}

u16 A1_rdw(u32 addr)
{
	u16 a = in16r((UWORD *)memaddr(addr));
	LOGMEM("rdw: %x -> %x\n", addr, a);
	return a;
}

u32 A1_rdl(u32 addr)
{
	u32 a = in32r((ULONG *)memaddr(addr));
	LOGMEM("rdl: %x -> %x\n", addr, a);
	return a;
}

void A1_wrb(u32 addr, u8 val)
{
	LOGMEM("wrb: %x <- %x\n", addr, val);
	out8((UBYTE *)memaddr(addr), val);
}

void A1_wrw(u32 addr, u16 val)
{
	LOGMEM("wrw: %x <- %x\n", addr, val);
	out16r((UWORD *)memaddr(addr), val);
}

void A1_wrl(u32 addr, u32 val)
{
	LOGMEM("wrl: %x <- %x\n", addr, val);
	out32r((ULONG *)memaddr(addr), val);
}

static X86EMU_memFuncs _A1_mem;

#define in_byte(from) in8( (UBYTE *)port_to_mem(from))
#define in_word(from) in16r((UWORD *)port_to_mem(from))
#define in_long(from) in32r((ULONG *)port_to_mem(from))
#define out_byte(to, val) out8((UBYTE *)port_to_mem(to), val)
#define out_word(to, val) out16r((UWORD *)port_to_mem(to), val)
#define out_long(to, val) out32r((ULONG *)port_to_mem(to), val)

u32 port_to_mem(int port)
{
	if (port >= 0xcfc && port <= 0xcff)
	    return 0xEEC00004; //0+port;
	else if (port >= 0xcf8 && port <= 0xcfb)
	    return 0xEEC00000; //+port;
	else
	    return CFG_ISA_IO_BASE_ADDRESS + port;
}

u8 A1_inb(int port)
{
	u8 a;
	a = in_byte(port);
	LOGIO(port, "inb: %Xh -> %d (%Xh)\n", port, a, a);
	return a;
}

u16 A1_inw(int port)
{
	u16 a = in_word(port);
	LOGIO(port, "inw: %Xh -> %d (%Xh)\n", port, a, a);
	return a;
}

u32 A1_inl(int port)
{
	u32 a = in_long(port);
	LOGIO(port, "inl: %Xh -> %d (%Xh)\n", port, a, a);
	return a;
}

void A1_outb(int port, u8 val)
{
	LOGIO(port, "outb: %Xh <- %d (%Xh)\n", port, val, val);
	out_byte(port, val);
}

void A1_outw(int port, u16 val)
{
	LOGIO(port, "outw: %Xh <- %d (%Xh)\n", port, val, val);
	out_word(port, val);
}

int blocked_port = 0;

void A1_outl(int port, u32 val)
{
	LOGIO(port, "outl: %Xh <- %d (%Xh)\n", port, val, val);

	// Workaround
	if (port != blocked_port)
		out_long(port, val);
	else
		LOGIO(port, "blocked\n");
}

static X86EMU_pioFuncs _A1_pio;

static int reloced_ops = 0;

void reloc_ops(void *reloc_addr)
{
	extern void (*x86emu_optab[256])(u8);
	extern void (*x86emu_optab2[256])(u8);
	extern void tables_relocate(unsigned int offset);
	int i;
	unsigned long delta;
	if (reloced_ops == 1) return;
	reloced_ops = 1;

	PRINTF("reloc_addr = %p\n", reloc_addr);
	delta = TEXT_BASE - (unsigned long)reloc_addr;
	PRINTF("delta = %p\n", delta);

	for (i=0; i<256; i++)
	{
		x86emu_optab[i] -= delta;
		x86emu_optab2[i] -= delta;
	}

	_A1_mem.rdb = A1_rdb;
	_A1_mem.rdw = A1_rdw;
	_A1_mem.rdl = A1_rdl;
	_A1_mem.wrb = A1_wrb;
	_A1_mem.wrw = A1_wrw;
	_A1_mem.wrl = A1_wrl;

	_A1_pio.inb = (u8 (X86APIP)(X86EMU_pioAddr))A1_inb;
	_A1_pio.inw = (u16 (X86APIP)(X86EMU_pioAddr))A1_inw;
	_A1_pio.inl = (u32 (X86APIP)(X86EMU_pioAddr))A1_inl;
	_A1_pio.outb = (void (X86APIP)(X86EMU_pioAddr, u8))A1_outb;
	_A1_pio.outw = (void (X86APIP)(X86EMU_pioAddr, u16))A1_outw;
	_A1_pio.outl = (void (X86APIP)(X86EMU_pioAddr, u32))A1_outl;

	tables_relocate(delta);

}

#define ANY_KEY(text)				\
	printf(text);				\
	while (!tstc());

unsigned char more_strap[] = {
		0xb4, 0x0, 0xb0, 0x2, 0xcd, 0x10,
};
#define MORE_STRAP_BYTES 6 // Additional bytes of strap code


unsigned char *done_msg="VGA Initialized\0";

int execute_bios(pci_dev_t gr_dev, void *reloc_addr)
{
	extern void bios_init(void);
	extern void remove_init_data(void);
	extern int video_rows(void);
	extern int video_cols(void);
	extern int video_size(int, int);
	u8 *strap;
	int i;
	char *pal_reset;

	PRINTF("Trying to remove init data\n");
	sam440_remove_init_data();
	PRINTF("Removed init data from cache, now in RAM\n");

	reloc_ops(reloc_addr);
	reloc_mode_table(reloc_addr);

	PRINTF("Attempting to run emulator on %02x:%02x:%02x\n",
	   PCI_BUS(gr_dev), PCI_DEV(gr_dev), PCI_FUNC(gr_dev));

	// Allocate memory
	// FIXME: We shouldn't use this much memory really.
	memset(&M, 0, sizeof(X86EMU_sysEnv));
	M.mem_base = (unsigned long)malloc(EMULATOR_MEM_SIZE);
	M.mem_size = (unsigned long)EMULATOR_MEM_SIZE;

	if (!M.mem_base)
	{
		PRINTF("Unable to allocate one megabyte for emulator\n");
		return 0;
	}

	if (attempt_map_rom(gr_dev, (void *)(M.mem_base + EMULATOR_BIOS_OFFSET)) == 0)
	{
		PRINTF("Error mapping rom. Emulation terminated\n");
		return 0;
	}

	strap = (u8*)M.mem_base + EMULATOR_STRAP_OFFSET;

	/*
	 * Poke the strap routine. This might need a bit of extending
	 * if there is a mode switch involved, i.e. we want to int10
	 * afterwards to set a different graphics mode, or alternatively
	 * there might be a different start address requirement if the
	 * ROM doesn't have an x86 image in its first image.
	 */

	PRINTF("Poking strap...\n");

	// FAR CALL c000:0003
	*strap++ = 0x9A; *strap++ = 0x03; *strap++ = 0x00;
	*strap++ = 0x00; *strap++ = 0xC0;

#if 1
	// insert additional strap code
	for (i=0; i < MORE_STRAP_BYTES; i++)
	{
		*strap++ = more_strap[i];
	}
#endif
	// HALT
	*strap++ = 0xF4;

	PRINTF("Done poking strap\n");

	/*
	 * Setup the init parameters.
	 * Per PCI specs, AH must contain the bus and AL
	 * must contain the devfn, encoded as (dev<<3)|fn
	 */

	PRINTF("Settingup init parameters\n");
	// Execution starts here
	M.x86.R_CS = SEG(EMULATOR_STRAP_OFFSET);
	M.x86.R_IP = OFF(EMULATOR_STRAP_OFFSET);

	// Stack at top of ram
	M.x86.R_SS = SEG(EMULATOR_STACK_OFFSET);
	M.x86.R_SP = OFF(EMULATOR_STACK_OFFSET);

	// Input parameters
	M.x86.R_AH = PCI_BUS(gr_dev);
	M.x86.R_AL = (PCI_DEV(gr_dev)<<3) | PCI_FUNC(gr_dev);

	PRINTF("Setting up I/O and memory access functions\n");
	// Set the I/O and memory access functions
	X86EMU_setupMemFuncs(&_A1_mem);
	PRINTF("PIO\n");
	X86EMU_setupPioFuncs(&_A1_pio);

	// If the initializing card is an ATI card, block access to port 0x34
	unsigned short vendor;
	pci_read_config_word(gr_dev, PCI_VENDOR_ID, &vendor);
	if (vendor == 0x1002)
	{
		PRINTF("Initializing a Radeon, blocking port access\n");
		int bar;

		for (bar = PCI_BASE_ADDRESS_0; bar <= PCI_BASE_ADDRESS_5; bar += 4)
		{
			unsigned int val;
			pci_read_config_dword(gr_dev, bar, &val);
			if (val & PCI_BASE_ADDRESS_SPACE_IO)
			{
				blocked_port = val & PCI_BASE_ADDRESS_IO_MASK;
				blocked_port += 0x34;
				break;
			}
		}
	}
	else
		blocked_port = 0;

	// Init the "BIOS".
	PRINTF("BIOS init\n");
	bios_init();
#if defined(DEBUG) && defined(SINGLESTEP)
#ifndef FORCE_SINGLESTEP
	s = _getenv("x86_singlestep");
	if (s && strcmp(s, "on")==0)
	{
#endif
		PRINTF("Enabling single stepping for debug\n");
		X86EMU_trace_on();
#ifndef FORCE_SINGLESTEP
	}
#endif
#endif

	// Ready set go...
	PRINTF("Running emulator\n");
	setup_tlb_for_cache(1);
	X86EMU_exec();
	setup_tlb_for_cache(0);

	PRINTF("Done running emulator\n");

/* FIXME: Remove me */
	pal_reset = getenv("x86_palette_reset");
	if (pal_reset && strcmp(pal_reset, "on") == 0)
	{
		PRINTF("Palette reset\n");

		out_byte(0x3C8, 0);
		out_byte(0x3C9, 0);
		out_byte(0x3C9, 0);
		out_byte(0x3C9, 0);
		for (i=0; i<254; i++)
		{
			out_byte(0x3C9, 63);
			out_byte(0x3C9, 63);
			out_byte(0x3C9, 63);
		}

		out_byte(0x3c0, 0x20);
	}
/* FIXME: remove me */
#ifdef EASTEREGG
	if (easteregg_active)
	{
		extern void video_easteregg(void);
		video_easteregg();
	}
#endif

	return 1;
}

// Clean up the x86 mess
void shutdown_bios(void)
{
//    disable_compatibility_hole();
	// Free the memory associated
//    free(M.mem_base);
//    setup_tlb_for_cache(0);
}

int to_int(char *buffer)
{
	int base = 0;
	int res  = 0;

	if (*buffer == '$')
	{
	base = 16;
	buffer++;
	}
	else base = 10;

	for (;;)
	{
	switch(*buffer)
	{
	case '0' ... '9':
	    res *= base;
	    res += *buffer - '0';
	    break;
	case 'A':
	case 'a':
	    res *= base;
	    res += 10;
	    break;
	case 'B':
	case 'b':
	    res *= base;
	    res += 11;
	    break;
	case 'C':
	case 'c':
	    res *= base;
	    res += 12;
	    break;
	case 'D':
	case 'd':
	    res *= base;
	    res += 13;
	    break;
	case 'E':
	case 'e':
	    res *= base;
	    res += 14;
	    break;
	case 'F':
	case 'f':
	    res *= base;
	    res += 15;
	    break;
	default:
	    return res;
	}
	buffer++;
	}
	return res;
}
